import logging
from os import PathLike
from typing import Union, List
from ruamel.yaml import YAML
from ruamel.yaml import CommentedMap
from .rules_parsing_helper import RulesParsingHelper
from ..handlers.rules_handler import RuleHandler
from .trigger_or_response_params import TriggerParams, ResponseParams


class RulesParser:

    def __init__(self):
        self.yaml = YAML()
        self.global_triggers = {}
        self.global_responses = {}
        self.trigger_params = TriggerParams(global_elements=self.global_triggers)
        self.response_params = ResponseParams(global_elements=self.global_responses)

    def parse_globals(self, config: CommentedMap) -> None:
        RulesParsingHelper.add_globals(config, self.trigger_params)
        RulesParsingHelper.add_globals(config, self.response_params)

    def parse_single_rule(self, rule: CommentedMap) -> RuleHandler:
        trigger = RulesParsingHelper.parse_rule_trigger_or_response(rule, self.trigger_params)
        response = RulesParsingHelper.parse_rule_trigger_or_response(rule, self.response_params)
        probability = rule['probability'] if 'probability' in rule else 1
        handler = RuleHandler(trigger, response, probability=probability)
        logging.info('Parsed %s', handler.handler_repr)
        return handler

    def parse_file(self, file_path: Union[PathLike, str, bytes]) -> List[RuleHandler]:
        with open(file_path, 'r', encoding='utf-8') as stream:
            config = self.yaml.load(stream)

        logging.info('Parsing globals from %s', file_path)
        self.parse_globals(config)

        logging.info('Parsing rules from %s', file_path)
        return [self.parse_single_rule(rule) for rule in config['rules']]



import logging
import os
from argparse import ArgumentParser
from telegram.ext import Updater
from .logging_config import configure_logger
from .paths import token_file_path, default_rules_file_path
from .configuration.rules_parser import RulesParser
from .handlers.combined_handlers import CombinedConversationHandler
from .handlers.default_error_handler import handle_error


def start_bot():
    """ Configures and runs the bot """
    configure_logger()

    parser = ArgumentParser(description='A rule-based Telegram bot engine, no coding required')
    parser.add_argument('--config', action='store', type=str, help='specify configuration file')
    args = parser.parse_args()
    logging.debug(f'Parsed arguments: {args}')

    with open(token_file_path(), 'r', encoding='utf-8') as token_file:
        bot_token = token_file.read().strip()

    rule_parser = RulesParser()

    if args.config:
        config_path = os.path.expanduser(os.path.expandvars(args.config))
        print(config_path)
        if os.access(config_path, os.R_OK):
            rules_file_path = config_path
        else:
            rules_file_path = default_rules_file_path()
            logging.warning(f'Cannot read {config_path}, defaulting to {rules_file_path}')
    else:
        rules_file_path = default_rules_file_path()

    logging.info('Reading and parsing rules file from %s', rules_file_path)
    rule_handlers = rule_parser.parse_file(rules_file_path)
    logging.info('Found %d rules', len(rule_handlers))
    conversation_handler = CombinedConversationHandler(rule_handlers)

    logging.info('Creating bot updater')
    updater = Updater(bot_token)
    updater.dispatcher.add_handler(conversation_handler)
    updater.dispatcher.add_error_handler(handle_error)
    updater.start_polling()

    # Run the bot until you press Ctrl-C or the process receives SIGINT,
    # SIGTERM or SIGABRT. This should be used most of the time, since
    # start_polling() is non-blocking and will stop the bot gracefully.
    logging.info('Moving the bot to idle mode to keep listening to updates')
    updater.idle()



from functools import reduce
from typing import List, Union, Optional
from telegram import Update
from telegram.ext import Filters, BaseFilter
from ..dict_enum import DictEnum
from .trigger_result import TriggerResult
from .basic_triggers import BaseTrigger
from ..users_helper import DEFAULT_USERS_HELPER


class FilterBasedTrigger(BaseTrigger):
    """
    A base class for filter-based triggers. This is used to trigger a rule when an incoming message
    passes through the specific filter.
    """

    def __init__(self, message_filter: BaseFilter):
        """
        Constructs the trigger.

        :param message_filter: The filter to test if the message passes through
        """
        super().__init__()
        self.filter = message_filter

    def check_trigger(self, update: Update) -> TriggerResult:
        return TriggerResult(should_respond=self.filter(update))


class _UserFilterBasedTrigger(FilterBasedTrigger):
    """
    A user filter trigger. This is used to trigger a rule when an incoming message comes from a
    specific user.
    """

    def __init__(
            self,
            nickname: Optional[str] = None,
            user_id: Optional[int] = None,
            username: Optional[str] = None,
    ):
        """
        Constructs the trigger.
        `nickname` can be used if such a nickname is defined in users.json file.
        Otherwise, one of `user_id` and `username` should be specified.

        :param message_filter: The filter to test if the message passes through
        :param nickname: The nickname of the user to pass messages from.
        :param user_id: The Telegram user ID of the user to pass messages from.
        :param username: The Telegram username of the user to pass messages from.
        """
        if nickname is not None:
            super().__init__(Filters.user(DEFAULT_USERS_HELPER.get_user_id_by_nickname(nickname)))
        else:
            super().__init__(Filters.user(user_id, username))


class _ChatFilterBasedTrigger(FilterBasedTrigger):
    """
    A chat filter trigger. This is used to trigger a rule when an incoming message is in a specific
    chat.
    """

    def __init__(self, chat_id: Optional[int] = None, username: Optional[str] = None):
        """
        Constructs the trigger.
        One and only one of `chat_id` and `username` should be specified.

        :param chat_id: The Telegram chat ID of the chat to pass messages from.
        :param username: The Telegram username whose chat to pass messages from.
        """
        super().__init__(Filters.chat(chat_id, username))


class _LanguageFilterBasedTrigger(FilterBasedTrigger):
    """
    A language filter trigger. This is used to trigger a rule when an incoming message is in a
    specific language.
    """

    def __init__(self, lang: Union[str, List[str]]):
        """
        Constructs the trigger.

        :param lang: The language code/s in which to pass messages.
        """
        super().__init__(Filters.language(lang))


class _MessageTypeFilterBasedTrigger(FilterBasedTrigger):
    """
    A message type filter trigger. This is used to trigger a rule when an incoming message is of a
    specific type.
    """

    def __init__(self, message_type):
        """
        Constructs the trigger.

        :param message_type: The message type of which to filter to pass messages, for example: \
            'photo', 'status_update.left_chat_member' or 'document'. See more in \
            :py:class:`telegram.ext.filters.Filters`.
        """
        subfilters = message_type.split('.')
        try:
            message_type_filter = reduce(getattr, subfilters, Filters)
            if isinstance(message_type_filter, BaseFilter):
                super().__init__(message_type_filter)
            else:
                raise ValueError(f'"{message_type}" is not a valid message type to filter by, '
                                 f'but it is a valid filter. Did you mean to use '
                                 f'"{message_type}" as a filter type instead?')
        except AttributeError:
            raise ValueError(  # pylint: disable=raise-missing-from
                f'"{message_type}" is not a valid message type to filter by.'
            )


class FilterTriggers(DictEnum):
    """Text-based triggers."""

    user = _UserFilterBasedTrigger
    """A user filter trigger. See more in :class:`_UserFilterBasedTrigger`."""

    chat = _ChatFilterBasedTrigger
    """A user filter trigger. See more in :class:`_ChatFilterBasedTrigger`."""

    language = _LanguageFilterBasedTrigger
    """A user filter trigger. See more in :class:`_LanguageFilterBasedTrigger`."""

    message_type = _MessageTypeFilterBasedTrigger
    """A user filter trigger. See more in :class:`_MessageTypeFilterBasedTrigger`."""



import logging


def configure_logger():
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(
        logging.Formatter(
            '%(asctime)s %(module)s.%(funcName)s %(levelname)s - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
        )
    )

    root_logger.addHandler(console_handler)



import logging
import random
from telegram import Update, Bot
from telegram.ext import CallbackContext
from ..responses.basic_responses import BaseResponse
from ..triggers.basic_triggers import BaseTrigger


class RuleHandler:
    """
    A handler for a single rule. Consists of a trigger checker (`BaseTrigger`),
    a responder (`BaseResponse`) and probability.
    Has a method `handle` that gets a Telegram bot and a Telegram update and responds as it should,
    if it should.
    """
    def __init__(self,
                 trigger_checker: BaseTrigger,
                 responder: BaseResponse,
                 probability: float = 1):
        self.trigger_checker = trigger_checker
        self.responder = responder

        self.handler_repr = f'{self.trigger_checker} -> {self.responder}'
        self.handler_repr = f'{self.handler_repr:70}'

        if probability > 1 or probability <= 0:
            raise ValueError(f'Parameter \'probability\' should be in range (0, 1], '
                             f'but {probability} was given')

        self.probability_to_respond = probability

    def handle(self, update: Update, context: CallbackContext):
        logging.debug('[%s] Received update %s', self.handler_repr, update.update_id)
        trigger_result = self.trigger_checker.check_trigger(update)
        if trigger_result.should_respond:
            logging.info('[%s] Bot should respond to update %s with a probability of %f',
                         self.handler_repr,
                         update.update_id,
                         self.probability_to_respond)
            if random.random() <= self.probability_to_respond:
                logging.info('[%s] Responding to update %s', self.handler_repr, update.update_id)
                self.responder.respond(context.bot, update, trigger_result.response_payload)



import json
import os
from typing import Optional
from .paths import users_file_path


class UsersHelper:
    """ A helper for loading and reading users from the users file """

    def __init__(self, file_path: Optional[os.PathLike | str | bytes] = None):
        if file_path is not None:
            self.file_path = file_path
        else:
            self.file_path = users_file_path()

        self._users = {}
        self.load_users()

    def load_users(self, force: bool = False):
        if force or not self._users:
            if os.path.exists(self.file_path):
                with open(self.file_path, 'r', encoding='utf-8') as file:
                    self._users = json.load(file)

    def get_user_id_by_nickname(self, nickname: str) -> int:
        return self._users[nickname]


DEFAULT_USERS_HELPER = UsersHelper()



from typing import Any
from telegram import Bot, Update
from telegram.message import Message


class ResponseHelper:
    @staticmethod
    def reply(bot: Bot, update: Update, response: Any, **kwargs: Any) -> Message:
        return bot.send_message(
            chat_id=update.effective_chat.id,
            text=response,
            reply_to_message_id=update.message.message_id,
            **kwargs,
        )

    @staticmethod
    def message(bot: Bot, update: Update, response: Any, **kwargs: Any) -> Message:
        return bot.send_message(chat_id=update.effective_chat.id, text=response, **kwargs)

    @staticmethod
    def document(bot: Bot, update: Update, response: Any, **kwargs: Any) -> Message:
        return bot.send_document(chat_id=update.effective_chat.id, document=response, **kwargs)



import abc
from typing import Dict, get_type_hints, Union, List, Any, Type
from inspect import isclass
from ruamel.yaml import CommentedMap, CommentedSeq
from .common_types import TriggerOrResponse, GlobalsDict
from .boolean_helper import BooleanHelper
from ..dict_enum import DictEnum


class TriggerOrResponseParser(abc.ABC):
    """ A common base parser for trigger/response configuration parsers"""

    def __init__(self, mapping_class: Type[DictEnum], element_base_class: Type[TriggerOrResponse]):
        self.__mapping_class = mapping_class
        self.__element_base_class = element_base_class

    @property
    def mapping_class(self) -> Type[DictEnum]:
        """ Returns the mapping class (`Triggers` or `Responses`)"""
        return self.__mapping_class

    @property
    def element_base_class(self) -> Type[TriggerOrResponse]:
        """ Returns the element base class (`BaseTrigger` or `BaseResponse`)"""
        return self.__element_base_class

    def parse_atomic(self, config: Dict[str, Any]) -> TriggerOrResponse:
        """
        Parses an atomic configuration subtree as a trigger/response
        :param config: A configuration subtree to parse
        :return: The trigger/response object created from the given configuration subtree
        """
        config_copy = dict(config)
        name = None
        if 'name' in config_copy:
            name = config_copy.pop('name')

        element = self.mapping_class[config_copy.pop('type')]

        # Some triggers (most of them) are classes and some are instances (mostly filter triggers).
        # This allows both cases to be used.
        # TODO: this comment does not seem to be correct any more. Try removing the case outside
        #  the "if" block; All triggers and responses seem to be classes.
        if isclass(element):
            trigger_or_response = element(**config_copy)
            if name:
                trigger_or_response.name = name
            return trigger_or_response
        return element

    def parse_single(self, config: Union[CommentedMap, str], global_elements: GlobalsDict) \
            -> TriggerOrResponse:
        """
        Recursively parses a single configuration subtree as a trigger/response
        :param config: A configuration subtree to parse
        :param global_elements: A dictionary with all triggers and responses configured globally
        :return: The trigger/response object created from the given configuration subtree
        """
        if isinstance(config, str):
            return BooleanHelper.parse_boolean_subrule(config, global_elements)

        config_copy = dict(config)
        parameters_to_parse = self.__find_parameters_to_parse_as_subelement(config)
        config.pop('type')

        for parameter in parameters_to_parse:
            config_copy[parameter] = self.parse_single(config[parameter], global_elements)

        return self.parse_atomic(config_copy)

    def parse_many(self, config: CommentedSeq, global_elements: GlobalsDict) \
            -> Dict[str, TriggerOrResponse]:
        """
        Parse all triggers/responses in a configuration
        :param config: A configuration subtree to config as a list of triggers/responses
        :param global_elements: A dictionary with all triggers and responses configured globally
        :return: A dictionary where keys are the original keys from the given configuration subtree
            and values are triggers/responses
        """
        return {
            element['name']: self.parse_single(element, global_elements)
            for element
            in config
        }

    def __find_parameters_to_parse_as_subelement(self, config: CommentedMap) -> List[str]:
        """
        Finds parameters in configuration which should be parsed as triggers/responses themselves
        (For example, when using a trigger/response that gets another trigger/response as a
        parameter).
        :param config: A configuration subtree to parse
        :return: List of the found parameter names
        """

        mapping_class_constructor = self.mapping_class[config['type']].__init__
        parameters_type_hints = get_type_hints(mapping_class_constructor)
        parameters_to_parse = []

        for parameter_name in config:

            # Meta-parameters to ignore
            if parameter_name in ['type', 'name']:
                continue

            # If this parameter has no type hint, we cannot check if it should be parsed or not
            if parameter_name not in parameters_type_hints:
                continue

            type_hint = parameters_type_hints[parameter_name]

            origin_type = getattr(type_hint, '__origin__', None)

            if origin_type is not None:
                if origin_type == Union:
                    if self.element_base_class in type_hint.__args__:
                        parameters_to_parse.append(parameter_name)
                else:
                    raise NotImplementedError(f'Origin type {origin_type} '
                                              f'is currently not supported')
            elif isclass(type_hint) and issubclass(type_hint, self.element_base_class):
                parameters_to_parse.append(parameter_name)

        return parameters_to_parse



from enum import Enum, EnumMeta


class DictEnumMeta(EnumMeta):
    """Metaclass for DictEnum"""

    def __getitem__(cls, key):
        key_path = key.split('.')
        item = super().__getitem__(key_path.pop(0)).value

        if key_path:
            return item['.'.join(key_path)]

        return item


class DictEnum(Enum, metaclass=DictEnumMeta):
    """
    A subscriptable enum class: An enum where elements can be accessed using the subscription
    operator.
    While regular enum values can only be accessed like `EnumName.value_name`, subscriptable enum
    values can also be access like `EnumName['value_name']`.
    The subscription key can be used to concatenate subscriptions, for example: `EnumName['x.y']`
    is equivalent to `EnumName['x']['y']`, as well as to `EnumName.x.y`.
    """



from typing import Callable
from boolean import boolean
from ruamel.yaml import CommentedMap
from .boolean_operators import OPERATOR_TO_FUNCTION
from .common_types import TriggerOrResponse, GlobalsDict


class BooleanHelper:
    """ A helper class for parsing boolean algebra expressions. """

    ParsingFunction = Callable[[CommentedMap, GlobalsDict], TriggerOrResponse]

    @staticmethod
    def evaluate_boolean_expression(expr: boolean.Expression,
                                    global_elements: GlobalsDict) -> TriggerOrResponse:
        """
        Convert a boolean expression into a `BaseTrigger`/`BaseResponse` object
        :param expr: A boolean expression parsed from a subrule (a configuration subtree)
        :param global_elements: A dictionary with all triggers and responses configured globally
        :return: The trigger/response object created from the given boolean expression
        """
        # If the trigger/response here is just a name, look for it in the globals
        if isinstance(expr, boolean.Symbol):
            return global_elements[str(expr)]

        boolean_function = OPERATOR_TO_FUNCTION[type(expr)]
        evaluated_args = [BooleanHelper.evaluate_boolean_expression(arg, global_elements)
                          for arg
                          in expr.args]
        return boolean_function(*evaluated_args)

    @staticmethod
    def parse_boolean_subrule(subrule: str, global_elements: GlobalsDict) -> TriggerOrResponse:
        """
        Convert a subrule (a configuration subtree) into a `BaseTrigger`/`BaseResponse` object
        :param subrule: A boolean subrule (for example, "rule1 and (not rule2)")
        :param global_elements: A dictionary with all triggers and responses configured globally
        :return: The trigger/response object created from the given subrule
        """
        algebra = boolean.BooleanAlgebra()
        parsed_expr = algebra.parse(subrule)
        return BooleanHelper.evaluate_boolean_expression(parsed_expr, global_elements)



from typing import Dict
from dataclasses import dataclass
from .common_types import GlobalsDict
from .trigger_or_response_parser import TriggerOrResponseParser
from ..triggers.basic_triggers import BaseTrigger
from ..responses.basic_responses import BaseResponse
from ..responses import Responses
from ..triggers import Triggers


@dataclass
class TriggerOrResponseParams:
    singular_key: str
    global_elements: GlobalsDict
    parser: TriggerOrResponseParser

    @property
    def plural_key(self):
        return self.singular_key + 's'


@dataclass
class TriggerParams(TriggerOrResponseParams):
    def __init__(self, global_elements: Dict[str, BaseTrigger]):
        super().__init__(singular_key='trigger',
                         global_elements=global_elements,
                         parser=TriggerOrResponseParser(Triggers, BaseTrigger))


@dataclass
class ResponseParams(TriggerOrResponseParams):
    def __init__(self, global_elements: Dict[str, BaseResponse]):
        super().__init__(singular_key='response',
                         global_elements=global_elements,
                         parser=TriggerOrResponseParser(Responses, BaseResponse))



from typing import Union, Dict
from ..triggers.basic_triggers import BaseTrigger
from ..responses.basic_responses import BaseResponse


TriggerOrResponse = Union[BaseTrigger, BaseResponse]
GlobalsDict = Dict[str, TriggerOrResponse]



import abc
from typing import Optional, Dict, Any

from telegram import Bot, Update
from telegram.message import Message
from ..representable import Representable


class BaseResponse(abc.ABC, Representable):
    def __init__(self, parse_mode: Optional[str] = None):
        super().__init__()
        self.__parse_mode = parse_mode

    @property
    def response_helper_kwargs(self) -> Dict[str, Any]:
        kwargs = {}

        if self.__parse_mode is not None:
            kwargs["parse_mode"] = self.__parse_mode

        return kwargs

    @abc.abstractmethod
    def respond(self, bot: Bot, update: Update, response_payload: dict) -> Message:
        pass



class Representable:
    """
    A "representable" interface for triggers and responses.
    The representation is the name of the trigger/response if given,
    or "inline <Type of trigger/response>" otherwise.
    """
    def __init__(self):
        self.__name = None

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, value):
        self.__name = value

    def __str__(self):
        if self.__name:
            return self.__name

        return f'inline {self.__class__.__name__}'
